import os
import psutil
import requests
import threading

from wafec_wrapt_custom.utility import fullname

base_url = os.environ.get('WWC_URL')
base_url = base_url if base_url else 'http://localhost:6543'


def _safe_run(target, args, kwargs):
    try:
        target(*args, **kwargs)
    except:
        pass


def _post_async(*args, **kwargs):
    thread = threading.Thread(target=_safe_run, args=(requests.post, args, kwargs))
    thread.start()


def add_proxy_interception_info(name, x=None, trace=None):
    p = psutil.Process()
    ps = p.name()
    data = {'ps': ps, 'name': name, 'x': fullname(x), 'trace': trace}
    _post_async(url=f'{base_url}/api/proxy/interception/add', json=data)



import wrapt
import abc
import copy
import json

from wafec_wrapt_custom.utility import fullname, starts_with
from wafec_wrapt_custom.comm.clients import add_proxy_interception_info

__all__ = [
    'WafecDefaultProxy',
    'create_proxy',
    'WafecDefaultProxyIterator',
    'add_to_white_list',
    'create_name'
]


_simple_types = (int, float, str, bool, complex)
_json_default_encoder_default = json._default_encoder.default
_white_list = ['nova.', 'glance.', 'neutron.', 'cinder.']

_default_add_proxy_interception_info = add_proxy_interception_info


def set_default_add_proxy_interception_info(new_func):
    global _default_add_proxy_interception_info
    _default_add_proxy_interception_info = new_func


def create_name(*args):
    name = ''
    for arg in [arg for arg in args if arg]:
        if type(arg) not in _simple_types and type(arg) not in (list, dict, tuple):
            name += fullname(arg) + ' '
        else:
            name += str(arg) + ' '
    return name.strip()


def add_to_white_list(item):
    global _white_list
    _white_list.append(item)


def _is_allowed(x):
    if not isinstance(x, wrapt.ObjectProxy):
        if isinstance(x, dict) or isinstance(x, list) or callable(x) or\
           type(x).__name__ == 'dict_items' or isinstance(x, tuple):
            return True
        if x is not None and type(x) not in _simple_types:
            if starts_with(_white_list, fullname(x)):
                return True
    return False


def _resolve_json_encoder():
    def default(cls, o):
        if isinstance(o, WafecDefaultProxy):
            return o.__wrapped__
        raise TypeError('')
    json._default_encoder.default = default.__get__(json._default_encoder, _json_default_encoder_default)


def create_proxy(x, trace=None):
    if _is_allowed(x):
        return WafecDefaultProxy(x, trace=trace)
    return x


class WafecDefaultProxy(wrapt.ObjectProxy, abc.ABC):
    def __init__(self, wrapped, trace=None):
        super(WafecDefaultProxy, self).__init__(wrapped)
        self._self_trace = trace

        _resolve_json_encoder()

    def __copy__(self):
        result = copy.copy(self.__wrapped__)
        return create_proxy(result)

    def __deepcopy__(self, memodict={}):
        result = copy.deepcopy(self.__wrapped__, memodict)
        return create_proxy(result)

    def __getitem__(self, item):
        trace = create_name(self._self_trace, f'[{item}]')
        _default_add_proxy_interception_info(item, x=self.__wrapped__, trace=trace)
        result = self.__wrapped__[item]
        return create_proxy(result, trace=trace)

    def __getattr__(self, item):
        trace = create_name(self._self_trace, f'.{item}')
        _default_add_proxy_interception_info(item, x=self.__wrapped__, trace=trace)
        result = getattr(self.__wrapped__, item)
        return create_proxy(result, trace=trace)

    def __call__(self, *args, **kwargs):
        result = self.__wrapped__(*args, **kwargs)
        return create_proxy(result, trace=self._self_trace)

    def __iter__(self):
        return WafecDefaultProxyIterator(self._self_trace, self.__wrapped__)

    def __reduce__(self):
        result = self.__wrapped__.__reduce__()
        return result

    def __reduce_ex__(self, protocol):
        result = self.__wrapped__.__reduce_ex__(protocol)
        return result


class WafecDefaultProxyIterator:
    def __init__(self, self_trace, wrapped):
        self.self_trace = self_trace
        self.wrapped = wrapped
        self.iterator = iter(wrapped)
        self.n = 0

    def __iter__(self):
        return self

    def __next__(self):
        result = next(self.iterator)
        self.n += 1
        trace = create_name(self.self_trace, f'[{self.n - 1}]')
        return create_proxy(result, trace=trace)



import unittest

from wafec_wrapt_custom.utility import fullname, starts_with, init_hierarchy
from wafec_wrapt_custom.tests.foo.bar import Bar


class FooBar(object):
    pass


class ClassWithAttribute(object):
    def __init__(self):
        self.arg1 = 'test1'
        self.arg2 = {'test2': 'value2'}
        self.arg3 = ['test3', 'test4']
        self.arg4 = ['test4', {'test5': 'value5'}]
        self.arg5 = 5


class UtilityTests(unittest.TestCase):
    def setUp(self):
        self.black_list = ['test1.', 'test2.test3.']

    def test_fullname_when_none(self):
        fn = fullname(None)
        self.assertIsNone(fn)

    def test_fullname(self):
        bar = Bar()
        fn = fullname(bar)
        self.assertEqual('wafec_wrapt_custom.tests.foo.bar.Bar', fn)

    def test_fullname_with_locals(self):
        foobar = FooBar()
        fn = fullname(foobar)
        self.assertEqual('utility_tests.FooBar', fn)

    def test_starts_with(self):
        result = starts_with(self.black_list, 'test1.Bar')
        self.assertTrue(result)

    def test_starts_with_when_not_present(self):
        result = starts_with(self.black_list, 'test3.Bar')
        self.assertFalse(result)

    def test_starts_with_when_none(self):
        result = starts_with(self.black_list, None)
        self.assertFalse(result)

    def test_starts_with_when_collection_is_none(self):
        result = starts_with(None, 'test3.Bar')
        self.assertFalse(result)



from flask import Flask, request, jsonify

app = Flask(__name__)

_FILE_NAME = './logs.txt'


@app.route('/')
def index():
    return 'Hello'


@app.route('/api/proxy/interception/add', methods=['POST'])
def add_proxy_interception_info():
    ps = request.json['ps']
    name = request.json['name']
    x = request.json['x']
    trace = request.trace['trace']

    with open(_FILE_NAME, 'a') as logger:
        logger.write(f'ps={ps}, name={name}, x={x}, trace={trace}\n')
    print(f'PROXY INFO ps={ps}, name={name}, x={x}, trace={trace}')
    return jsonify({'success': True})


def start_server():
    app.run(host="0.0.0.0", port=6543)


if __name__ == '__main__':
    start_server()



import argparse

from wafec_wrapt_custom.comm.servers import start_server
from wafec_wrapt_custom.comm.clients import add_proxy_interception_info


def define_clients_add_proxy_interception_info(subparsers):
    def defaults(args):
        add_proxy_interception_info(args.name)
    parser = subparsers.add_parser('add_proxy_interception_info')
    parser.add_argument('-n', '--name', dest='name', type=str, required=True)
    parser.set_defaults(func=defaults)


def define_clients(subparsers):
    parser = subparsers.add_parser('clients')
    subparsers = parser.add_subparsers()
    define_clients_add_proxy_interception_info(subparsers)


def define_servers_start(subparsers):
    def defaults(args):
        start_server()

    parser = subparsers.add_parser('start')
    parser.set_defaults(func=defaults)


def define_servers(subparsers):
    parser = subparsers.add_parser('servers')
    subparsers = parser.add_subparsers()
    define_servers_start(subparsers)


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    define_servers(subparsers)
    define_clients(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == '__main__':
    main()





__all__ = [
    'fullname',
    'starts_with'
]


# https://stackoverflow.com/questions/2020014/get-fully-qualified-class-name-of-an-object-in-python
def fullname(o):
    if o is None:
        return None
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    else:
        return module + '.' + o.__class__.__name__


def starts_with(collection, text):
    if text is None or collection is None:
        return False
    for item in collection:
        if text.startswith(item):
            return True
    return False



from setuptools import setup, find_packages


setup(
    name="wafec.wrapt.custom",
    version="1.0.14",
    author="Wallace",
    author_email="wallacefcardoso@gmail.com",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'wrapt>=1.12.1',
        'Flask>=1.1.2',
        'requests>=2.0.0',
        'psutil>=5.7.0'
    ],
    entry_points={
        'console_scripts': [
            'comm_cmd = wafec_wrapt_custom.comm_command_line:main'
        ]
    }
)



import requests
import json
import logging

LOG = logging.getLogger(__name__)


class DataEvent(object):
    def __init__(self, source, event_name, value, data_type, is_callable):
        self.source = str(source)
        self.event_name = str(event_name)
        self.value = str(value)
        self.data_type = str(data_type)
        self.is_callable = is_callable
        self.process_name = 'unset'

    def to_dict(self):
        return {
            'source': self.source,
            'event_name': self.event_name,
            'value': self.value,
            'is_callable': self.is_callable,
            'data_type': self.data_type,
            'process_name': self.process_name
        }


class DataManagerAdapter(object):
    def notify(self, data_event):
        pass


class DataManager(object):
    def __init__(self, api_uri):
        self._api_uri = api_uri

    def notify(self, data_event):
        try:
            data = data_event.to_dict()
            headers = {
                'Content-Type': 'application/json'
            }
            requests.post(self._api_uri, data=json.dumps(data), headers=headers)
        except Exception as exc:
            LOG.exception(str(exc))



import unittest
import wrapt
import copy
import json
import pickle

from wafec_wrapt_custom.proxies import *
from wafec_wrapt_custom.proxies import _is_allowed, _white_list, set_default_add_proxy_interception_info, fullname,\
    _default_add_proxy_interception_info


class ProxiesTests(unittest.TestCase):
    def setUp(self):
        def custom_callable():
            pass

        self.callable_x = custom_callable
        self.inst_x = CustomTestClass('test1', 'description1')

        add_to_white_list('proxies_tests.')

    def test_is_allowed_when_dict(self):
        res = _is_allowed({})
        self.assertTrue(res)

    def test_is_allowed_when_list(self):
        res = _is_allowed([])
        self.assertTrue(res)

    def test_is_allowed_when_tuple(self):
        res = _is_allowed((1,))
        self.assertTrue(res)

    def test_is_allowed_when_callable(self):
        res = _is_allowed(self.callable_x)
        self.assertTrue(res)

    def test_is_allowed_when_obj(self):
        res = _is_allowed(CustomTestClass('name1', 'description1'))
        self.assertTrue(res)

    def test_is_allowed_when_int(self):
        res = _is_allowed(1)
        self.assertFalse(res)

    def test_is_allowed_when_bool(self):
        res = _is_allowed(True)
        self.assertFalse(res)

    def test_is_allowed_when_float(self):
        res = _is_allowed(10.0)
        self.assertFalse(res)

    def test_is_allowed_when_complex(self):
        res = _is_allowed(2j)
        self.assertFalse(res)

    def test_is_allowed_when_text(self):
        res = _is_allowed('test1')
        self.assertFalse(res)

    def test_is_allowed_when_none(self):
        res = _is_allowed(None)
        self.assertFalse(res)

    def test_create_name_when_obj(self):
        name = create_name(self.inst_x)
        self.assertEqual('proxies_tests.CustomTestClass', name)

    def test_create_name_when_string(self):
        name = create_name('test')
        self.assertEqual('test', name)

    def test_create_name_when_number(self):
        name = create_name(10)
        self.assertEqual('10', name)

    def test_create_name_when_list(self):
        name = create_name([1, 2])
        self.assertEqual('[1, 2]', name)

    def test_create_name_when_dict(self):
        name = create_name({'test': 'value'})
        self.assertEqual("{'test': 'value'}", name)

    def test_create_name_with_combination_of_obj_and_text(self):
        name = create_name(self.inst_x, 'test')
        self.assertEqual('proxies_tests.CustomTestClass test', name)


class WafecDefaultProxyTests(unittest.TestCase):
    def setUp(self):
        self.dict_x = dict()
        self.dict_x['test1'] = 'value1'
        self.dict_x['test2'] = 'value2'
        self.dict_x['test3'] = dict()
        self.dict_x['test3']['test31'] = 'value31'

        self.list_x = list()
        self.list_x.append('test1')
        self.list_x.append('test2')
        self.list_x.append(list())
        self.list_x[2].append(dict())
        self.list_x[2][0]['test3'] = 'value3'

        self.inst_x = CustomTestClass('name1', 'description1')

        _white_list.append('proxies_tests.')

    def test_create_proxy_with_dict(self):
        d = create_proxy(self.dict_x)
        self.assertIsInstance(d, wrapt.ObjectProxy)
        self.assertIsInstance(d, dict)
        self.assertEqual('value1', d['test1'])
        self.assertNotIsInstance(d['test1'], wrapt.ObjectProxy)

    def test_create_proxy_with_dict_when_keys_is_used(self):
        d = create_proxy(self.dict_x)
        for k1, k2 in zip(d.keys(), self.dict_x.keys()):
            self.assertEqual(k2, k1)
            self.assertNotIsInstance(k1, wrapt.ObjectProxy)

    def test_create_proxy_with_dict_when_getting_proxy(self):
        d = create_proxy(self.dict_x)
        self.assertIsInstance(d['test3'], wrapt.ObjectProxy)

    def test_create_proxy_with_dict_when_items_is_used(self):
        d = create_proxy(self.dict_x)
        for i, item1, item2 in zip(range(len(self.dict_x.keys())), self.dict_x.items(), d.items()):
            self.assertEqual(item1[0], item2[0])
            self.assertEqual(item1[1], item2[1])
            if i == 2:
                self.assertIsInstance(item2[1], WafecDefaultProxy)
            else:
                self.assertIsNot(item2[1], WafecDefaultProxy)
            self.assertNotIsInstance(item2[0], WafecDefaultProxy)

    def test_create_proxy_with_list(self):
        l = create_proxy(self.list_x)
        self.assertIsInstance(l, WafecDefaultProxy)

    def test_create_proxy_with_list_when_getting_text(self):
        l = create_proxy(self.list_x)
        text = l[0]
        self.assertNotIsInstance(text, WafecDefaultProxy)

    def test_create_proxy_with_list_when_getting_other_list(self):
        l = create_proxy(self.list_x)
        other_list = l[2]
        self.assertIsInstance(other_list, WafecDefaultProxy)

    def test_create_proxy_with_list_when_getting_dict(self):
        l = create_proxy(self.list_x)
        other_list = l[2]
        d = other_list[0]
        self.assertIsInstance(d, WafecDefaultProxy)
        self.assertNotIsInstance(d['test3'], WafecDefaultProxy)

    def test_create_proxy_with_elementary_when_int(self):
        n = create_proxy(int(10))
        self.assertIsInstance(n, int)
        self.assertNotIsInstance(n, WafecDefaultProxy)

    def test_create_proxy_with_elementary_when_bool(self):
        b = create_proxy(bool(True))
        self.assertIsInstance(b, bool)
        self.assertNotIsInstance(b, WafecDefaultProxy)

    def test_create_proxy_with_elementary_when_float(self):
        f = create_proxy(float(10.0))
        self.assertIsInstance(f, float)
        self.assertNotIsInstance(f, WafecDefaultProxy)

    def test_create_proxy_with_elementary_when_text(self):
        d = create_proxy(str('test'))
        self.assertIsInstance(d, str)
        self.assertNotIsInstance(d, WafecDefaultProxy)

    def test_create_proxy_with_obj(self):
        inst = create_proxy(self.inst_x)
        self.assertIsInstance(inst, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_getting_text(self):
        inst = create_proxy(self.inst_x)
        text = inst.name
        self.assertEqual('name1', text)
        self.assertNotIsInstance(text, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_getting_int(self):
        inst = create_proxy(self.inst_x)
        id = inst.id
        self.assertEqual(1, id)
        self.assertNotIsInstance(id, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_getting_dict(self):
        inst = create_proxy(self.inst_x)
        d = inst.data
        self.assertIsInstance(d, dict)
        self.assertIsInstance(d, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_getting_list(self):
        inst = create_proxy(self.inst_x)
        l = inst.collection
        self.assertIsInstance(l, list)
        self.assertIsInstance(l, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_callable_returns_text(self):
        inst = create_proxy(self.inst_x)
        c = inst.get_name
        text = c()
        self.assertTrue(callable(c))
        self.assertIsInstance(c, WafecDefaultProxy)
        self.assertNotIsInstance(text, WafecDefaultProxy)
        self.assertEqual(text, 'name1')

    def test_create_proxy_with_obj_when_callable_returns_dict(self):
        inst = create_proxy(self.inst_x)
        c = inst.get_data
        d = c()
        self.assertTrue(callable(c))
        self.assertIsInstance(d, dict)
        self.assertIsInstance(d, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_property_returns_text(self):
        inst = create_proxy(self.inst_x)
        text = inst.name_description
        self.assertNotIsInstance(text, WafecDefaultProxy)

    def test_create_proxy_with_obj_when_property_returns_data(self):
        inst = create_proxy(self.inst_x)
        data = inst.more_data
        self.assertIsInstance(data, WafecDefaultProxy)

    def test_create_proxy_with_dict_and_copy(self):
        d = create_proxy(self.dict_x)
        c = copy.copy(d)
        self.assertIsInstance(c, WafecDefaultProxy)

    def test_create_proxy_with_dict_and_deep_copy(self):
        d = create_proxy(self.dict_x)
        c = copy.deepcopy(d)
        self.assertIsInstance(c, WafecDefaultProxy)

    def test_create_proxy_with_list_and_copy(self):
        l = create_proxy(self.list_x)
        c = copy.copy(l)
        self.assertIsInstance(c, WafecDefaultProxy)

    def test_create_proxy_with_list_and_deep_copy(self):
        l = create_proxy(self.list_x)
        c = copy.deepcopy(l)
        self.assertIsInstance(c, WafecDefaultProxy)

    def test_create_proxy_with_dict_and_json_dumps(self):
        d = create_proxy({'test': 'value'})
        j = json.dumps(d)
        self.assertEqual('{"test": "value"}', j)

    def test_create_proxy_with_dict_and_pickle_dumps(self):
        o = {'test': 'value'}
        d = create_proxy(o)
        d_pickle_string = pickle.dumps(d)
        o_pickle_string = pickle.dumps(o)
        o_x = pickle.loads(o_pickle_string)
        d_x = pickle.loads(d_pickle_string)
        self.assertEqual(o_x['test'], d_x['test'])

    def test_create_proxy_with_obj_and_pickle_dumps(self):
        pickle_string = pickle.dumps(create_proxy(self.inst_x))
        inst = pickle.loads(pickle_string)
        self.assertIsInstance(inst, CustomTestClass)

    def test_create_proxy_with_list_when_iterable_is_used(self):
        l = create_proxy(self.list_x)
        for item in l:
            self.assertIsNotNone(item)

    def test_create_proxy_with_list_when_getting_iterable(self):
        l = create_proxy(self.list_x)
        it = iter(l)
        self.assertIsInstance(it, WafecDefaultProxyIterator)
        self.assertTrue(hasattr(it, '__iter__'))

    def test_create_proxy_with_name(self):
        o = create_proxy(self.inst_x, 'test1')
        self.assertEqual('test1', o._self_trace)

    def test_create_proxy_with_name_when_dict(self):
        d = create_proxy(self.dict_x, 'test1')
        d3 = d['test3']
        self.assertEqual('test1 [test3]', d3._self_trace)

    def test_create_proxy_with_name_when_dict_and_first_name_is_not_provided(self):
        d = create_proxy(self.dict_x)
        d3 = d['test3']
        self.assertEqual('[test3]', d3._self_trace)

    def test_create_proxy_with_name_when_obj_is_returned_from_method(self):
        inst = create_proxy(self.inst_x, trace='test')
        result = inst.get_data()
        self.assertEqual('test .get_data', result._self_trace)


class WafecDefaultProxyInterceptionTests(unittest.TestCase):
    def setUp(self):
        self._original = _default_add_proxy_interception_info

        self.inst_x = CustomTestClass('test1', 'description1')

    def tearDown(self):
        set_default_add_proxy_interception_info(self._original)

    def test_create_proxy_and_add_interception(self):
        def wrapper(item, x, trace):
            self.assertEqual('get_data', item)
            self.assertIsInstance(x, CustomTestClass)
            self.assertEqual('proxies_tests.CustomTestClass', fullname(x))
            self.assertEqual('test .get_data', trace)
        set_default_add_proxy_interception_info(wrapper)
        d = create_proxy(self.inst_x, 'test')
        data = d.get_data()
        self.assertIsNotNone(data)

    def test_create_proxy_and_add_interception_when_method_returns(self):
        def wrapper(item, x, trace):
            self.assertEqual('test .get_data [test1]', trace)
        d = create_proxy(self.inst_x, 'test')
        data = d.get_data()
        data['test1'] = 'value1'
        set_default_add_proxy_interception_info(wrapper)
        value1 = data['test1']
        self.assertEqual('value1', value1)


class CustomTestClassBase(object):
    def __init__(self, name):
        self.name = name


class CustomTestClass(CustomTestClassBase):
    def __init__(self, name, description):
        CustomTestClassBase.__init__(self, name)
        self.description = description
        self.data = dict()
        self.collection = list()
        self.id = int(1)

    def get_name(self):
        return self.name

    def set_name(self, name):
        self.name = name

    def get_data(self):
        return self.data

    @property
    def name_description(self):
        return f'{self.name} {self.description}'

    @property
    def more_data(self):
        return self.data
